/*
 -------------------------------------------------------------------------
 Copyright (c) 2005, XIMETA, Inc., IRVINE, CA, USA.
 All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary 
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright 
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products 
      built using this software without specific written permission. 
      
 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.
 
 DISCLAIMER

 This software is provided 'as is' with no explcit or implied warranties
 in respect of any properties, including, but not limited to, correctness 
 and fitness for purpose.
 -------------------------------------------------------------------------
*/
#ifndef _SAL_MEM_H_
#define _SAL_MEM_H_

#include "sal/types.h"
#include "sal/sal.h"
/**
 * memory block.
 * 
 * @param ptr the pointer to the data
 * @param private the pointer to the private reference
 * @param len the lenth of the data pointed by ptr
 */
struct sal_mem_block {
    xpointer ptr;
    xpointer private;
    xsize_t len;    
};

#ifdef DEBUG_USE_MEM_TAG
NDAS_SAL_API extern void *sal_malloc_tag(unsigned int size, const char* name, int line);
NDAS_SAL_API extern void sal_free_tag(void *ptr);
#define sal_malloc(size)    sal_malloc_tag(size, __FUNCTION__, __LINE__)
#define sal_free(ptr)        sal_free_tag(ptr)
#else
/* Allocate the memroy */
NDAS_SAL_API extern void    *sal_malloc(unsigned int size);
/**
 * Free the memory
 */
NDAS_SAL_API extern void sal_free(void *ptr);


#endif
NDAS_SAL_API extern void sal_mem_display(int verbose);

#ifdef DEBUG_MEMORY_LEAK
#ifndef DEBUG_USE_MEM_TAG
#error DEBUG_USE_MEM_TAG should be defined
#endif
#define sal_alloc_from_pool(pool, size) sal_malloc(size)
#define sal_free_from_pool(pool, ptr) sal_free(ptr)
#else
NDAS_SAL_API extern void* sal_alloc_from_pool(void* pool, int size); 
NDAS_SAL_API extern void sal_free_from_pool(void* pool, void* ptr); 
#endif
/* used only when USE_SYSTEM_MEM_POOL is defined */
NDAS_SAL_API extern void* sal_create_mem_pool(char* name, int size); 
NDAS_SAL_API extern int sal_destroy_mem_pool(void* pool); 

#endif /* _SAL_MEM_H_ */
