/*
 -------------------------------------------------------------------------
 Copyright (c) 2005, XIMETA, Inc., IRVINE, CA, USA.
 All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary 
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright 
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products 
      built using this software without specific written permission. 
      
 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.
 
 DISCLAIMER

 This software is provided 'as is' with no explcit or implied warranties
 in respect of any properties, including, but not limited to, correctness 
 and fitness for purpose.
 -------------------------------------------------------------------------
*/
/*
    Abstracts synchronization mechanism of operating system

 */

#ifndef _SAL_SYNC_H
#define _SAL_SYNC_H
#include "sal/sal.h"
#include "sal/time.h"

#define SAL_SYNC_FOREVER    (-1)
#define SAL_SYNC_NOWAIT        0

#define SAL_INVALID_MUTEX    (sal_mutex)(0)
#define SAL_INVALID_SEMAPHORE (sal_semaphore)(0)

#define SAL_SYNC_OK                     (0)
#define SAL_SYNC_ERR                    (-1)
#define SAL_SYNC_TIMEOUT            (-2)
#define SAL_SYNC_INTERRUPTED    (-3)

struct _sal_mutex;
typedef struct _sal_mutex* sal_mutex;
struct _sal_semaphore;
typedef struct _sal_semaphore* sal_semaphore;

#if defined(XPLAT_ENABLE_EMBED_LINUX_OPT)

#elif defined(UCOSII)
#include <ucos_ii.h>
#define sal_mutex_create(desc) (sal_mutex)(1)
#define sal_mutex_destroy(m) do { } while(0)
#define sal_mutex_take(m,f) ({ OSDisableInt(); f; })
#define sal_mutex_give(m,f) ({ OSEnableInt(); f; })
#else
/* Mutex
    Only the thread that locked the mutex can unlock the mutex 
*/
NDAS_SAL_API sal_mutex     sal_mutex_create(const char *desc);
NDAS_SAL_API void        sal_mutex_destroy(sal_mutex m);
NDAS_SAL_API void        sal_mutex_take(sal_mutex m, unsigned long* flags);
NDAS_SAL_API void        sal_mutex_give(sal_mutex m, unsigned long* flags);
#endif 

NDAS_SAL_API sal_semaphore    sal_semaphore_create(const char *desc, int max_count, int initial_count);
NDAS_SAL_API void        sal_semaphore_destroy(sal_semaphore b);

/* Return 
    SAL_SYNC_OK for successful semaphore taking 
    SAL_SYNC_TIMEOUT for timeout
    SAL_SYNC_INTERRUPTED if wait function is interrupted by signal
*/
NDAS_SAL_API int        sal_semaphore_take(sal_semaphore b, sal_tick tick);
NDAS_SAL_API int        sal_semaphore_give(sal_semaphore b);

NDAS_SAL_API int        sal_semaphore_getvalue(sal_semaphore b);

#endif    /* !_SAL_SYNC_H */

