/*
 -------------------------------------------------------------------------
 Copyright (c) 2002-2006, XIMETA, Inc., FREMONT, CA, USA.
 All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary 
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright 
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer#include "sal/io.h"
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products 
      built using this software without specific written permission. 
      
 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.
 
 DISCLAIMER

 This software is provided 'as is' with no explcit or implied warranties
 in respect of any properties, including, but not limited to, correctness 
 and fitness for purpose.
 -------------------------------------------------------------------------
*/
#include <linux/version.h> // LINUX_VERSION
#include <linux/module.h> // EXPORT_SYMBOL
#include <linux/spinlock.h> // spinklock_t
#include <linux/sched.h> // jiffies
#include <asm/semaphore.h> // struct semaphore
#include <asm/atomic.h> // atomic
#include <linux/mm.h> //
#include <linux/slab.h> // 2.4 kmalloc/kfree
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/ide.h>
#include <linux/smp_lock.h>
#include <linux/time.h>

#include "sal/sync.h"
#include "sal/thread.h"
#include "sal/time.h"
#include "sal/mem.h"
#include "sal/debug.h"
#include "sal/types.h"
#include "sal/io.h"

#include "linux_ver.h" // SAL_HZ

NDAS_SAL_API sal_file sal_file_open(const char * filename, int flags, int mode)
{
	int linux_flags = 0;
	sal_file nfile;
	if (flags & SAL_FILE_FLAG_RO) 
		linux_flags |= O_RDONLY;
	if (flags & SAL_FILE_FLAG_RW)
		linux_flags |= O_RDWR;
	nfile = (sal_file) filp_open(filename, linux_flags, mode);
	if (nfile == NULL)
		printk("SAL: Failed to open file %s\n", filename);
	return nfile;
}
EXPORT_SYMBOL(sal_file_open);

NDAS_SAL_API xbool     sal_file_get_size(sal_file file, xuint64* size)
{
	struct file* filp = (struct file*) file;
	mm_segment_t oldfs;
	int	error = ENOTTY;
	xbool ret;
#if !(LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10))
	int size32;
#endif

	//
	// Assume file is always block device file.	
	//
	lock_kernel();

	if (filp->f_op && filp->f_op->ioctl) {
		oldfs = get_fs();
		set_fs(get_ds());
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10))
		error = filp->f_op->ioctl(filp->f_dentry->d_inode, filp, BLKGETSIZE64, (int)size);
#else
		error = filp->f_op->ioctl(filp->f_dentry->d_inode, filp, BLKGETSIZE, (int)&size32);
		*size = size32<<9;
#endif
//		printk("SAL: File size = %lld\n", *size);
		smp_mb();
		set_fs(oldfs);
		if (error)
			ret = FALSE;
		else
			ret = TRUE;
	} else {
		ret = FALSE;
	}
	unlock_kernel();
	return ret;
}

EXPORT_SYMBOL(sal_file_get_size);

NDAS_SAL_API void     sal_file_close(sal_file file)
{
	struct file* filp = (struct file*) file;
	filp_close(filp, NULL);
}
EXPORT_SYMBOL(sal_file_close);

NDAS_SAL_API xint32     sal_file_read(sal_file file, xuchar* buf, xint32 size, xuint64 offset)
{
	struct file* filp = (struct file*) file;
	mm_segment_t oldfs;
	int retval;
	loff_t foffset = offset;
//	printk("SAL: pos=%llx, read size = %d\n", filp->f_pos, size);
	oldfs = get_fs();
	set_fs(get_ds());
#if 1
	retval = filp->f_op->read(filp, buf, size, &foffset);
#else
	// read throughput test. do nothing when read...
	retval = size;
#endif
	smp_mb();
	set_fs(oldfs);	
	if (retval !=size) 
		return -1;
	else
		return size;
}
EXPORT_SYMBOL(sal_file_read);

NDAS_SAL_API xint32     sal_file_write(sal_file file, xuchar* buf, xint32 size, xuint64 offset)
{
	struct file* filp = (struct file*) file;
	mm_segment_t oldfs;
	int retval;
	loff_t foffset = offset;
	
//	printk("SAL: pos=%llx, write size = %d\n", filp->f_pos, size);
	oldfs = get_fs();
	set_fs(get_ds());
#if 1	
	retval = filp->f_op->write(filp, buf, size, &foffset);
#else
	// write throughput test. do nothing when read...
	retval = size;
#endif
	smp_mb();
	set_fs(oldfs);
	if (retval !=size) 
		return -1;
	else
		return size;
}
EXPORT_SYMBOL(sal_file_write);

NDAS_SAL_API xuint64     sal_file_lseek(sal_file file, xuint64 offset, int whence)
{
	struct file* filp = (struct file*) file;
	mm_segment_t oldfs;
	oldfs = get_fs();
	set_fs(get_ds());

	switch(whence) {
	case SAL_SEEK_SET:
		filp->f_pos = offset;
		break;
	case SAL_SEEK_CUR:
		filp->f_pos += offset;
		break;
	case SAL_SEEK_END:
		// not implemented;
		return -1;
	}

	smp_mb();
	set_fs(oldfs);
	return filp->f_pos;
}
EXPORT_SYMBOL(sal_file_lseek);

